import numpy
import pytest
import warnings

import cupy
from cupy import testing
from cupy.exceptions import ComplexWarning


class TestBasic:
    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty(self, xp, dtype, order):
        a = xp.empty((2, 3, 4), dtype=dtype, order=order)
        a.fill(0)
        return a

    @testing.slow
    def test_empty_huge_size(self):
        a = cupy.empty((1024, 2048, 1024), dtype='b')
        a.fill(123)
        assert (a == 123).all()
        # Free huge memory for slow test
        del a
        cupy.get_default_memory_pool().free_all_blocks()

    @testing.slow
    def test_empty_huge_size_fill0(self):
        a = cupy.empty((1024, 2048, 1024), dtype='b')
        a.fill(0)
        assert (a == 0).all()
        # Free huge memory for slow test
        del a
        cupy.get_default_memory_pool().free_all_blocks()

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_scalar(self, xp, dtype, order):
        a = xp.empty((), dtype=dtype, order=order)
        a.fill(0)
        return a

    @testing.with_requires('numpy>=1.20', 'numpy<2.3')
    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_scalar_none(self, xp, dtype, order):
        with testing.assert_warns(DeprecationWarning):
            a = xp.empty(None, dtype=dtype, order=order)
        a.fill(0)
        return a

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_int(self, xp, dtype, order):
        a = xp.empty(3, dtype=dtype, order=order)
        a.fill(0)
        return a

    @testing.slow
    def test_empty_int_huge_size(self):
        a = cupy.empty(2 ** 31, dtype='b')
        a.fill(123)
        assert (a == 123).all()
        # Free huge memory for slow test
        del a
        cupy.get_default_memory_pool().free_all_blocks()

    @testing.slow
    def test_empty_int_huge_size_fill0(self):
        a = cupy.empty(2 ** 31, dtype='b')
        a.fill(0)
        assert (a == 0).all()
        # Free huge memory for slow test
        del a
        cupy.get_default_memory_pool().free_all_blocks()

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.empty_like(a, order=order)
        b.fill(0)
        return b

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like_contiguity(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.empty_like(a, order=order)
        b.fill(0)
        if order in ['f', 'F']:
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
        return b

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like_contiguity2(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        a = xp.asfortranarray(a)
        b = xp.empty_like(a, order=order)
        b.fill(0)
        if order in ['c', 'C']:
            assert b.flags.c_contiguous
        else:
            assert b.flags.f_contiguous
        return b

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like_contiguity3(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        # test strides that are both non-contiguous and non-descending
        a = a[:, ::2, :].swapaxes(0, 1)
        b = xp.empty_like(a, order=order)
        b.fill(0)
        if order in ['k', 'K', None]:
            assert not b.flags.c_contiguous
            assert not b.flags.f_contiguous
        elif order in ['f', 'F']:
            assert not b.flags.c_contiguous
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
            assert not b.flags.f_contiguous
        return b

    @testing.for_all_dtypes()
    def test_empty_like_K_strides(self, dtype):
        # test strides that are both non-contiguous and non-descending
        a = testing.shaped_arange((2, 3, 4), numpy, dtype)
        a = a[:, ::2, :].swapaxes(0, 1)
        b = numpy.empty_like(a, order='K')
        b.fill(0)

        # GPU case
        ag = testing.shaped_arange((2, 3, 4), cupy, dtype)
        ag = ag[:, ::2, :].swapaxes(0, 1)
        bg = cupy.empty_like(ag, order='K')
        bg.fill(0)

        # make sure NumPy and CuPy strides agree
        assert b.strides == bg.strides
        return

    @testing.with_requires('numpy>=1.19')
    @testing.for_all_dtypes()
    def test_empty_like_invalid_order(self, dtype):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp, dtype)
            with pytest.raises(ValueError):
                xp.empty_like(a, order='Q')

    def test_empty_like_subok(self):
        a = testing.shaped_arange((2, 3, 4), cupy)
        with pytest.raises(TypeError):
            cupy.empty_like(a, subok=True)

    @testing.for_CF_orders()
    @testing.with_requires('numpy>=1.23')
    def test_empty_zero_sized_array_strides(self, order):
        a = numpy.empty((1, 0, 2), dtype='d', order=order)
        b = cupy.empty((1, 0, 2), dtype='d', order=order)
        assert b.strides == a.strides

    @pytest.mark.parametrize('offset', [1, -1, 1 << 63, -(1 << 63)])
    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_eye(self, xp, dtype, order, offset):
        return xp.eye(5, 4, offset, dtype, order=order)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_identity(self, xp, dtype):
        return xp.identity(4, dtype)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_zeros(self, xp, dtype, order):
        return xp.zeros((2, 3, 4), dtype=dtype, order=order)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_zeros_scalar(self, xp, dtype, order):
        return xp.zeros((), dtype=dtype, order=order)

    @testing.with_requires('numpy>=1.20', 'numpy<2.3')
    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_zeros_scalar_none(self, xp, dtype, order):
        with testing.assert_warns(DeprecationWarning):
            return xp.zeros(None, dtype=dtype, order=order)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_zeros_int(self, xp, dtype, order):
        return xp.zeros(3, dtype=dtype, order=order)

    @testing.for_CF_orders()
    def test_zeros_strides(self, order):
        a = numpy.zeros((2, 3), dtype='d', order=order)
        b = cupy.zeros((2, 3), dtype='d', order=order)
        assert b.strides == a.strides

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_zeros_like(self, xp, dtype, order):
        a = xp.ndarray((2, 3, 4), dtype=dtype)
        return xp.zeros_like(a, order=order)

    def test_zeros_like_subok(self):
        a = cupy.ndarray((2, 3, 4))
        with pytest.raises(TypeError):
            cupy.zeros_like(a, subok=True)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_ones(self, xp, dtype, order):
        return xp.ones((2, 3, 4), dtype=dtype, order=order)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_ones_like(self, xp, dtype, order):
        a = xp.ndarray((2, 3, 4), dtype=dtype)
        return xp.ones_like(a, order=order)

    def test_ones_like_subok(self):
        a = cupy.ndarray((2, 3, 4))
        with pytest.raises(TypeError):
            cupy.ones_like(a, subok=True)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_full(self, xp, dtype, order):
        return xp.full((2, 3, 4), 1, dtype=dtype, order=order)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_full_default_dtype(self, xp, dtype, order):
        return xp.full((2, 3, 4), xp.array(1, dtype=dtype), order=order)

    @testing.for_all_dtypes_combination(('dtype1', 'dtype2'))
    @testing.numpy_cupy_array_equal()
    @testing.with_requires('numpy<2.0')
    def test_full_dtypes_cpu_input(self, xp, dtype1, dtype2):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', ComplexWarning)
            return xp.full(
                (2, 3, 4), numpy.array(1, dtype=dtype1), dtype=dtype2)

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_full_like(self, xp, dtype, order):
        a = xp.ndarray((2, 3, 4), dtype=dtype)
        return xp.full_like(a, 1, order=order)

    @testing.for_all_dtypes_combination(('dtype1', 'dtype2'))
    @testing.numpy_cupy_array_equal()
    def test_full_like_dtypes_cpu_input(self, xp, dtype1, dtype2):
        a = xp.ndarray((2, 3, 4), dtype=dtype1)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', ComplexWarning)
            return xp.full_like(a, numpy.array(1, dtype=dtype1))

    def test_full_like_subok(self):
        a = cupy.ndarray((2, 3, 4))
        with pytest.raises(TypeError):
            cupy.full_like(a, 1, subok=True)


@testing.parameterize(
    *testing.product({
        'shape': [4, (4, ), (4, 2), (4, 2, 3), (5, 4, 2, 3)],
    })
)
class TestBasicReshape:

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like_reshape(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.empty_like(a, order=order, shape=self.shape)
        b.fill(0)
        return b

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    def test_empty_like_reshape_cupy_only(self, dtype, order):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.empty_like(a, shape=self.shape)
        b.fill(0)
        c = cupy.empty(self.shape, order=order, dtype=dtype)
        c.fill(0)

        testing.assert_array_equal(b, c)

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like_reshape_contiguity(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        b = xp.empty_like(a, order=order, shape=self.shape)
        b.fill(0)
        if order in ['f', 'F']:
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
        return b

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    def test_empty_like_reshape_contiguity_cupy_only(self, dtype, order):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.empty_like(a, order=order, shape=self.shape)
        b.fill(0)
        c = cupy.empty(self.shape)
        c.fill(0)
        if order in ['f', 'F']:
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
        testing.assert_array_equal(b, c)

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like_reshape_contiguity2(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        a = xp.asfortranarray(a)
        b = xp.empty_like(a, order=order, shape=self.shape)
        b.fill(0)
        shape = self.shape if not numpy.isscalar(self.shape) else (self.shape,)
        if (order in ['c', 'C'] or
                (order in ['k', 'K', None] and len(shape) != a.ndim)):
            assert b.flags.c_contiguous
        else:
            assert b.flags.f_contiguous
        return b

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    def test_empty_like_reshape_contiguity2_cupy_only(self, dtype, order):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        a = cupy.asfortranarray(a)
        b = cupy.empty_like(a, order=order, shape=self.shape)
        b.fill(0)
        c = cupy.empty(self.shape)
        c.fill(0)
        shape = self.shape if not numpy.isscalar(self.shape) else (self.shape,)
        if (order in ['c', 'C'] or
                (order in ['k', 'K', None] and len(shape) != a.ndim)):
            assert b.flags.c_contiguous
        else:
            assert b.flags.f_contiguous
        testing.assert_array_equal(b, c)

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_empty_like_reshape_contiguity3(self, xp, dtype, order):
        a = testing.shaped_arange((2, 3, 4), xp, dtype)
        # test strides that are both non-contiguous and non-descending
        a = a[:, ::2, :].swapaxes(0, 1)
        b = xp.empty_like(a, order=order, shape=self.shape)
        b.fill(0)
        shape = self.shape if not numpy.isscalar(self.shape) else (self.shape,)
        if len(shape) == 1:
            assert b.flags.c_contiguous
            assert b.flags.f_contiguous
        elif order in ['k', 'K', None] and len(shape) == a.ndim:
            assert not b.flags.c_contiguous
            assert not b.flags.f_contiguous
        elif order in ['f', 'F']:
            assert not b.flags.c_contiguous
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
            assert not b.flags.f_contiguous
        return b

    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    def test_empty_like_reshape_contiguity3_cupy_only(self, dtype, order):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        # test strides that are both non-contiguous and non-descending
        a = a[:, ::2, :].swapaxes(0, 1)
        b = cupy.empty_like(a, order=order, shape=self.shape)
        b.fill(0)
        shape = self.shape if not numpy.isscalar(self.shape) else (self.shape,)
        if len(shape) == 1:
            assert b.flags.c_contiguous
            assert b.flags.f_contiguous
        elif order in ['k', 'K', None] and len(shape) == a.ndim:
            assert not b.flags.c_contiguous
            assert not b.flags.f_contiguous
        elif order in ['f', 'F']:
            assert not b.flags.c_contiguous
            assert b.flags.f_contiguous
        else:
            assert b.flags.c_contiguous
            assert not b.flags.f_contiguous

        c = cupy.zeros(self.shape)
        c.fill(0)
        testing.assert_array_equal(b, c)

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_all_dtypes()
    def test_empty_like_K_strides_reshape(self, dtype):
        # test strides that are both non-contiguous and non-descending
        a = testing.shaped_arange((2, 3, 4), numpy, dtype)
        a = a[:, ::2, :].swapaxes(0, 1)
        b = numpy.empty_like(a, order='K', shape=self.shape)
        b.fill(0)

        # GPU case
        ag = testing.shaped_arange((2, 3, 4), cupy, dtype)
        ag = ag[:, ::2, :].swapaxes(0, 1)
        bg = cupy.empty_like(ag, order='K', shape=self.shape)
        bg.fill(0)

        # make sure NumPy and CuPy strides agree
        assert b.strides == bg.strides
        return

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_zeros_like_reshape(self, xp, dtype, order):
        a = xp.ndarray((2, 3, 4), dtype=dtype)
        return xp.zeros_like(a, order=order, shape=self.shape)

    @testing.for_CF_orders()
    @testing.for_all_dtypes()
    def test_zeros_like_reshape_cupy_only(self, dtype, order):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.zeros_like(a, shape=self.shape)
        c = cupy.zeros(self.shape, order=order, dtype=dtype)

        testing.assert_array_equal(b, c)

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_ones_like_reshape(self, xp, dtype, order):
        a = xp.ndarray((2, 3, 4), dtype=dtype)
        return xp.ones_like(a, order=order, shape=self.shape)

    @testing.for_all_dtypes()
    def test_ones_like_reshape_cupy_only(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.ones_like(a, shape=self.shape)
        c = cupy.ones(self.shape, dtype=dtype)

        testing.assert_array_equal(b, c)

    @testing.with_requires('numpy>=1.17.0')
    @testing.for_orders('CFAK')
    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_full_like_reshape(self, xp, dtype, order):
        a = xp.ndarray((2, 3, 4), dtype=dtype)
        return xp.full_like(a, 1, order=order, shape=self.shape)

    @testing.for_all_dtypes()
    def test_full_like_reshape_cupy_only(self, dtype):
        a = testing.shaped_arange((2, 3, 4), cupy, dtype)
        b = cupy.full_like(a, 1, shape=self.shape)
        c = cupy.full(self.shape, 1, dtype=dtype)

        testing.assert_array_equal(b, c)



import numpy

from cupy import cuda
from cupy._creation.basic import _new_like_order_and_strides
from cupy._core import internal


def _update_shape(a, shape):
    if shape is None and a is not None:
        shape = a.shape
    elif isinstance(shape, int):
        shape = (shape,)
    else:
        shape = tuple(shape)
    return shape


def empty_pinned(shape, dtype=float, order='C'):
    """Returns a new, uninitialized NumPy array with the given shape
    and dtype.

    This is a convenience function which is just :func:`numpy.empty`,
    except that the underlying memory is pinned/pagelocked.

    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        dtype: Data type specifier.
        order ({'C', 'F'}): Row-major (C-style) or column-major
            (Fortran-style) order.

    Returns:
        numpy.ndarray: A new array with elements not initialized.

    .. seealso:: :func:`numpy.empty`

    """
    shape = _update_shape(None, shape)
    nbytes = internal.prod(shape) * numpy.dtype(dtype).itemsize
    mem = cuda.alloc_pinned_memory(nbytes)
    out = numpy.ndarray(shape, dtype=dtype, buffer=mem, order=order)
    return out


def empty_like_pinned(a, dtype=None, order='K', subok=None, shape=None):
    """Returns a new, uninitialized NumPy array with the same shape and dtype
    as those of the given array.

    This is a convenience function which is just :func:`numpy.empty_like`,
    except that the underlying memory is pinned/pagelocked.

    This function currently does not support ``subok`` option.

    Args:
        a (numpy.ndarray or cupy.ndarray): Base array.
        dtype: Data type specifier. The data type of ``a`` is used by default.
        order ({'C', 'F', 'A', or 'K'}): Overrides the memory layout of the
            result. ``'C'`` means C-order, ``'F'`` means F-order, ``'A'`` means
            ``'F'`` if ``a`` is Fortran contiguous, ``'C'`` otherwise.
            ``'K'`` means match the layout of ``a`` as closely as possible.
        subok: Not supported yet, must be None.
        shape (int or tuple of ints): Overrides the shape of the result. If
            ``order='K'`` and the number of dimensions is unchanged, will try
            to keep order, otherwise, ``order='C'`` is implied.

    Returns:
        numpy.ndarray: A new array with same shape and dtype of ``a`` with
        elements not initialized.

    .. seealso:: :func:`numpy.empty_like`

    """
    # We're kinda duplicating the code here because order='K' needs special
    # treatment: strides need to be computed
    if subok is not None:
        raise TypeError('subok is not supported yet')
    if dtype is None:
        dtype = a.dtype
    shape = _update_shape(a, shape)
    order, strides, _ = _new_like_order_and_strides(
        a, dtype, order, shape, get_memptr=False)
    nbytes = internal.prod(shape) * numpy.dtype(dtype).itemsize
    mem = cuda.alloc_pinned_memory(nbytes)
    out = numpy.ndarray(shape, dtype=dtype, buffer=mem,
                        strides=strides, order=order)
    return out


def zeros_pinned(shape, dtype=float, order='C'):
    """Returns a new, zero-initialized NumPy array with the given shape
    and dtype.

    This is a convenience function which is just :func:`numpy.zeros`,
    except that the underlying memory is pinned/pagelocked.

    Args:
        shape (int or tuple of ints): Dimensionalities of the array.
        dtype: Data type specifier.
        order ({'C', 'F'}): Row-major (C-style) or column-major
            (Fortran-style) order.

    Returns:
        numpy.ndarray: An array filled with zeros.

    .. seealso:: :func:`numpy.zeros`

    """
    out = empty_pinned(shape, dtype, order)
    numpy.copyto(out, 0, casting='unsafe')
    return out


def zeros_like_pinned(a, dtype=None, order='K', subok=None, shape=None):
    """Returns a new, zero-initialized NumPy array with the same shape and dtype
    as those of the given array.

    This is a convenience function which is just :func:`numpy.zeros_like`,
    except that the underlying memory is pinned/pagelocked.

    This function currently does not support ``subok`` option.

    Args:
        a (numpy.ndarray or cupy.ndarray): Base array.
        dtype: Data type specifier. The dtype of ``a`` is used by default.
        order ({'C', 'F', 'A', or 'K'}): Overrides the memory layout of the
            result. ``'C'`` means C-order, ``'F'`` means F-order, ``'A'`` means
            ``'F'`` if ``a`` is Fortran contiguous, ``'C'`` otherwise.
            ``'K'`` means match the layout of ``a`` as closely as possible.
        subok: Not supported yet, must be None.
        shape (int or tuple of ints): Overrides the shape of the result. If
            ``order='K'`` and the number of dimensions is unchanged, will try
            to keep order, otherwise, ``order='C'`` is implied.

    Returns:
        numpy.ndarray: An array filled with zeros.

    .. seealso:: :func:`numpy.zeros_like`

    """  # NOQA
    out = empty_like_pinned(a, dtype, order, subok, shape)
    numpy.copyto(out, 0, casting='unsafe')
    return out



import pytest

import cupy
from cupy.cuda import driver
from cupy.cuda import runtime
from cupy import testing
import cupyx.scipy.signal  # NOQA

import numpy as np

try:
    import scipy  # NOQA
    import scipy.signal  # NOQA
except ImportError:
    pass


def _gen_gaussians(xp, center_locs, sigmas, total_length):
    xdata = xp.arange(0, total_length).astype(float)
    out_data = xp.zeros(total_length, dtype=float)
    for ind, sigma in enumerate(sigmas):
        tmp = (xdata - center_locs[ind]) / sigma
        out_data += xp.exp(-(tmp**2))
    return out_data


def _gen_gaussians_even(xp, sigmas, total_length):
    num_peaks = len(sigmas)
    delta = total_length / (num_peaks + 1)
    center_locs = xp.linspace(
        delta, total_length - delta, num=num_peaks).astype(int)
    out_data = _gen_gaussians(xp, center_locs, sigmas, total_length)
    return out_data, center_locs


@pytest.mark.xfail(
    runtime.is_hip and driver.get_build_version() < 5_00_00000,
    reason='name_expressions with ROCm 4.3 may not work')
@testing.with_requires('scipy')
class TestPeakProminences:

    @pytest.mark.parametrize('x', [[1, 2, 3], []])
    @testing.numpy_cupy_allclose(scipy_name="scp", type_check=False)
    def test_empty(self, x, xp, scp):
        """
        Test if an empty array is returned if no peaks are provided.
        """
        out = scp.signal.peak_prominences(x, [])
        return out

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, xp, scp):
        """
        Test if height of prominences is correctly calculated in signal with
        rising baseline (peak widths are 1 sample).
        """
        # Prepare basic signal
        x = xp.array([-1, 1.2, 1.2, 1, 3.2, 1.3, 2.88, 2.1])
        peaks = xp.array([1, 2, 4, 6])
        # Test if calculation matches handcrafted result
        out = scp.signal.peak_prominences(x, peaks)
        return out

    @pytest.mark.parametrize(
        'x', [[0.0, 2, 1, 2, 1, 2, 0], [0, 1.0, 0, 1, 0, 1, 0]])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_edge_cases(self, x, xp, scp):
        """
        Test edge cases.
        """
        # Peaks have same height, prominence and bases
        x = xp.asarray(x)
        peaks = xp.asarray([1, 3, 5])
        out = scp.signal.peak_prominences(x, peaks)
        return out

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_non_contiguous(self, xp, scp):
        """
        Test with non-C-contiguous input arrays.
        """
        x = xp.repeat(xp.asarray([-9, 9, 9, 0, 3, 1.0]), 2)
        peaks = xp.repeat(xp.asarray([1, 2, 4]), 2)
        out = scp.signal.peak_prominences(x[::2], peaks[::2])
        return out

    @testing.with_requires("numpy<2.0")
    @pytest.mark.parametrize('wlen', [8, 7, 6, 5, 3.2, 3, 1.1])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_wlen(self, wlen, xp, scp):
        """
        Test if wlen actually shrinks the evaluation range correctly.
        """
        x = xp.asarray([0, 1, 2, 3.0, 1, 0, -1])
        peak = xp.asarray([3])
        return scp.signal.peak_prominences(x, peak, wlen)

    def test_exceptions(self):
        """
        Verify that exceptions and warnings are raised.
        """
        for xp, scp in [(cupy, cupyx.scipy), (np, scipy)]:
            # x with dimension > 1
            with pytest.raises(ValueError, match='1-D array'):
                scp.signal.peak_prominences([[0, 1, 1, 0]], [1, 2])
            # peaks with dimension > 1
            with pytest.raises(ValueError, match='1-D array'):
                scp.signal.peak_prominences([0, 1, 1, 0], [[1, 2]])
            # x with dimension < 1
            with pytest.raises(ValueError, match='1-D array'):
                scp.signal.peak_prominences(3, [0,])

            # empty x with supplied
            with pytest.raises(ValueError, match='not a valid index'):
                scp.signal.peak_prominences([], [0])
            # invalid indices with non-empty x
            for p in [-100, -1, 3, 1000]:
                with pytest.raises(ValueError, match='not a valid index'):
                    scp.signal.peak_prominences([1, 0, 2], [p])

            # wlen < 3
            with pytest.raises(ValueError, match='wlen'):
                scp.signal.peak_prominences(xp.arange(10), [3, 5], wlen=1)


@pytest.mark.xfail(
    runtime.is_hip and driver.get_build_version() < 5_00_00000,
    reason='name_expressions with ROCm 4.3 may not work')
@testing.with_requires('scipy')
class TestPeakWidths:

    @pytest.mark.parametrize('x', [[1, 2, 3], []])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_empty(self, x, xp, scp):
        """
        Test if an empty array is returned if no peaks are provided.
        """
        widths = scp.signal.peak_widths(x, [])
        return widths

    @pytest.mark.filterwarnings("ignore:some peaks have a width of 0")
    @pytest.mark.parametrize('rel_height', [0, 0.25, 0.5, 0.75, 1.0, 2.0, 3.0])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, rel_height, xp, scp):
        """
        Test a simple use case with easy to verify results at different
        relative heights.
        """
        x = xp.array([1, 0, 1, 2, 1, 0, -1])
        out = scp.signal.peak_widths(x, [3], rel_height)
        return out

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_non_contiguous(self, xp, scp):
        """
        Test with non-C-contiguous input arrays.
        """
        x = xp.repeat(xp.asarray([0, 100, 50]), 4)
        peaks = xp.repeat(xp.asarray([1]), 3)
        result = scp.signal.peak_widths(x[::4], peaks[::3])
        return result

    def test_exceptions(self):
        """
        Verify that argument validation works as intended.
        """
        for xp, scp in [(cupy, cupyx.scipy), (np, scipy)]:
            with pytest.raises(ValueError, match='1-D array'):
                # x with dimension > 1
                scp.signal.peak_widths(xp.zeros((3, 4)), xp.ones(3))
            with pytest.raises(ValueError, match='1-D array'):
                # x with dimension < 1
                scp.signal.peak_widths(3, [0])
            with pytest.raises(ValueError, match='1-D array'):
                # peaks with dimension > 1
                scp.signal.peak_widths(
                    xp.arange(10), xp.ones((3, 2), dtype=xp.intp))
            with pytest.raises(ValueError, match='1-D array'):
                # peaks with dimension < 1
                scp.signal.peak_widths(xp.arange(10), 3)
            with pytest.raises(ValueError, match='not a valid index'):
                # peak pos exceeds x.size
                scp.signal.peak_widths(xp.arange(10), [8, 11])
            with pytest.raises(ValueError, match='not a valid index'):
                # empty x with peaks supplied
                scp.signal.peak_widths([], [1, 2])
            with pytest.raises(ValueError, match='rel_height'):
                # rel_height is < 0
                scp.signal.peak_widths([0, 1, 0, 1, 0], [1, 3], rel_height=-1)
            with pytest.raises(TypeError, match='None'):
                # prominence data contains None
                scp.signal.peak_widths(
                    [1, 2, 1], [1], prominence_data=(None, None, None))

    def test_mismatching_prominence_data(self):
        """Test with mismatching peak and / or prominence data."""
        for xp, scp in [(cupy, cupyx.scipy), (np, scipy)]:
            x = xp.asarray([0, 1, 0])
            peak = [1]
            for i, (prominences, left_bases, right_bases) in enumerate([
                ((1.,), (-1,), (2,)),  # left base not in x
                ((1.,), (0,), (3,)),  # right base not in x
                ((1.,), (2,), (0,)),  # swapped bases same as peak
                ((1., 1.), (0, 0), (2, 2)),  # array shapes don't match peaks
                ((1., 1.), (0,), (2,)),  # arrays with different shapes
                ((1.,), (0, 0), (2,)),  # arrays with different shapes
                ((1.,), (0,), (2, 2))  # arrays with different shapes
            ]):
                # Make sure input is matches output of signal.peak_prominences
                prominence_data = (xp.array(prominences, dtype=xp.float64),
                                   xp.array(left_bases, dtype=xp.intp),
                                   xp.array(right_bases, dtype=xp.intp))
                # Test for correct exception
                if i < 3:
                    match = "prominence data is invalid"
                else:
                    match = ("arrays in `prominence_data` must have the "
                             "same shape")
                with pytest.raises(ValueError, match=match):
                    scp.signal.peak_widths(
                        x, peak, prominence_data=prominence_data)

    @pytest.mark.filterwarnings("ignore:some peaks have a width of 0")
    @pytest.mark.parametrize('rel_height', [0, 2/3])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_intersection_rules(self, rel_height, xp, scp):
        """Test if x == eval_height counts as an intersection."""
        # Flatt peak with two possible intersection points if evaluated at 1
        x = [0, 1, 2, 1, 3, 3, 3, 1, 2, 1, 0]
        # relative height is 0 -> width is 0 as well, raises warning
        out = scp.signal.peak_widths(x, peaks=[5], rel_height=rel_height)
        return out


@pytest.mark.xfail(
    runtime.is_hip and driver.get_build_version() < 5_00_00000,
    reason='name_expressions with ROCm 4.3 may not work')
@testing.with_requires('scipy')
class TestFindPeaks:

    # Keys of optionally returned properties
    property_keys = {'peak_heights', 'left_thresholds', 'right_thresholds',
                     'prominences', 'left_bases', 'right_bases', 'widths',
                     'width_heights', 'left_ips', 'right_ips'}

    @testing.numpy_cupy_allclose(scipy_name="scp", type_check=False)
    def test_constant(self, xp, scp):
        """
        Test behavior for signal without local maxima.
        """
        open_interval = (None, None)
        peaks, props = scp.signal.find_peaks(
            xp.ones(10), height=open_interval, threshold=open_interval,
            prominence=open_interval, width=open_interval)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @pytest.mark.parametrize(
        'plateau_size', [(None, None), 4, (None, 3.5), (5, 50)])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_plateau_size(self, plateau_size, xp, scp):
        """
        Test plateau size condition for peaks.
        """
        # Prepare signal with peaks with peak_height == plateau_size
        plateau_sizes = xp.array([1, 2, 3, 4, 8, 20, 111])
        x = xp.zeros(plateau_sizes.size * 2 + 1)
        x[1::2] = plateau_sizes
        repeats = xp.ones(x.size, dtype=int)
        repeats[1::2] = x[1::2]
        x = xp.repeat(x, repeats.tolist())

        # Test full output
        peaks, props = scp.signal.find_peaks(x, plateau_size=plateau_size)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @pytest.mark.parametrize(
        'height', [(None, None), 0.5, (None, 3), (2, 3)])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_height_condition(self, height, xp, scp):
        """
        Test height condition for peaks.
        """
        x = xp.asarray([0., 1/3, 0., 2.5, 0, 4., 0])
        peaks, props = scp.signal.find_peaks(x, height=height)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @pytest.mark.parametrize(
        'threshold', [(None, None), 2, 3.5, (None, 5), (None, 4), (2, 4)])
    @testing.numpy_cupy_allclose(scipy_name="scp", type_check=False)
    def test_threshold_condition(self, threshold, xp, scp):
        """
        Test threshold condition for peaks.
        """
        x = xp.asarray([0, 2, 1, 4, -1])
        peaks, props = scp.signal.find_peaks(x, threshold=threshold)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @pytest.mark.parametrize('distance', [3, 3.0001])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_distance_condition(self, distance, xp, scp):
        """
        Test distance condition for peaks.
        """
        # Peaks of different height with constant distance 3
        peaks_all = xp.arange(1, 21, 3)
        x = xp.zeros(21)
        x[peaks_all] += xp.linspace(1, 2, peaks_all.size)
        peaks, props = scp.signal.find_peaks(x, distance=distance)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_distance_priority(self, xp, scp):
        # Test priority of peak removal
        x = xp.asarray([-2, 1, -1, 0, -3])
        # use distance > x size
        peaks, props = scp.signal.find_peaks(x, distance=10)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_prominence_condition(self, xp, scp):
        """
        Test prominence condition for peaks.
        """
        x = xp.linspace(0, 10, 100)
        peaks_true = xp.arange(1, 99, 2)
        offset = xp.linspace(1, 10, peaks_true.size)
        x[peaks_true] += offset
        interval = (3, 9)

        peaks, props = scp.signal.find_peaks(x, prominence=interval)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @testing.numpy_cupy_allclose(scipy_name="scp", type_check=False)
    def test_width_condition(self, xp, scp):
        """
        Test width condition for peaks.
        """
        x = xp.array([1, 0, 1, 2, 1, 0, -1, 4, 0])
        peaks, props = scp.signal.find_peaks(
            x, width=(None, 2), rel_height=0.75)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_properties(self, xp, scp):
        """
        Test returned properties.
        """
        open_interval = (None, None)
        x = xp.asarray([0, 1, 0, 2, 1.5, 0, 3, 0, 5, 9])
        peaks, props = scp.signal.find_peaks(
            x, height=open_interval, threshold=open_interval,
            prominence=open_interval, width=open_interval)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])

    def test_raises(self):
        """
        Test exceptions raised by function.
        """
        for xp, scp in [(cupy, cupyx.scipy), (np, scipy)]:
            with pytest.raises(ValueError, match="1-D array"):
                scp.signal.find_peaks(xp.array(1))
            with pytest.raises(ValueError, match="1-D array"):
                scp.signal.find_peaks(xp.ones((2, 2)))
            with pytest.raises(ValueError, match="distance"):
                scp.signal.find_peaks(xp.arange(10), distance=-1)

    @testing.with_requires("numpy<2.0")
    @pytest.mark.filterwarnings("ignore:some peaks have a prominence of 0",
                                "ignore:some peaks have a width of 0")
    @testing.numpy_cupy_allclose(scipy_name="scp", type_check=False)
    def test_wlen_smaller_plateau(self, xp, scp):
        """
        Test behavior of prominence and width calculation if the given window
        length is smaller than a peak's plateau size.

        Regression test for gh-9110.
        """
        peaks, props = scp.signal.find_peaks(
            [0, 1, 1, 1, 0], prominence=(None, None),
            width=(None, None), wlen=2)
        return (peaks,) + tuple(
            [props[k] for k in self.property_keys if k in props])


@pytest.mark.xfail(
    runtime.is_hip and driver.get_build_version() < 5_00_00000,
    reason='name_expressions with ROCm 4.3 may not work')
@testing.with_requires('scipy')
class TestArgrel:

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_empty(self, xp, scp):
        # When there are no relative extrema, make sure that
        # the number of empty arrays returned matches the
        # dimension of the input.
        z1 = xp.zeros(5)
        i = scp.signal.argrelmin(z1)

        z2 = xp.zeros((3, 5))
        row1, col1 = scp.signal.argrelmin(z2, axis=0)
        row2, col2 = scp.signal.argrelmin(z2, axis=1)
        return i[0], row1, col1, row2, col2

    @pytest.mark.parametrize('func_name', ['argrelmax', 'argrelmin'])
    @pytest.mark.parametrize('axis', [0, 1])
    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_basic(self, func_name, axis, xp, scp):
        # Note: the docstrings for the argrel{min,max,extrema} functions
        # do not give a guarantee of the order of the indices, so we'll
        # sort them before testing.

        x = xp.array([[1, 2, 2, 3, 2],
                      [2, 1, 2, 2, 3],
                      [3, 2, 1, 2, 2],
                      [2, 3, 2, 1, 2],
                      [1, 2, 3, 2, 1]])

        func = getattr(scp.signal, func_name)
        row, col = func(x, axis=axis)
        order = xp.argsort(row)
        return row[order], col[order]

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_highorder(self, xp, scp):
        order = 2
        sigmas = [1.0, 2.0, 10.0, 5.0, 15.0]
        test_data, act_locs = _gen_gaussians_even(xp, sigmas, 500)
        test_data[act_locs + order] = test_data[act_locs]*0.99999
        test_data[act_locs - order] = test_data[act_locs]*0.99999
        rel_max_locs = scp.signal.argrelmax(
            test_data, order=order, mode='clip')[0]
        return rel_max_locs

    @testing.numpy_cupy_allclose(scipy_name="scp")
    def test_2d_gaussians(self, xp, scp):
        sigmas = [1.0, 2.0, 10.0]
        test_data, _ = _gen_gaussians_even(xp, sigmas, 100)
        rot_factor = 20
        rot_range = xp.arange(0, len(test_data)) - rot_factor
        test_data_2 = xp.vstack([test_data, test_data[rot_range]])
        rel_max_rows, rel_max_cols = scp.signal.argrelmax(
            test_data_2, axis=1, order=1)
        return rel_max_rows, rel_max_cols



"""
The source code here is an adaptation with minimal changes from the following
SciPy Cython file:

https://github.com/scipy/scipy/blob/master/scipy/special/sph_harm.pxd
"""

from cupy import _core

from cupyx.scipy.special._poch import poch_definition
from cupyx.scipy.special._lpmv import lpmv_definition

sph_harmonic_definition = (
    poch_definition
    + lpmv_definition
    + """

#include <cupy/complex.cuh>

// include for CUDART_NAN, CUDART_INF
#include <cupy/math_constants.h>

#define NPY_PI        3.141592653589793238462643383279502884  /* pi */

// from scipy/special/sph_harm.pxd
__device__ complex<double> sph_harmonic(int m, int n, double theta, double phi)
{
    double x, prefactor;
    complex<double> val;
    int mp;

    x = cos(phi);
    if (abs(m) > n)
    {
        // sf_error.error("sph_harm", sf_error.ARG,
        //                "m should not be greater than n")
        return CUDART_NAN;
    }
    if (n < 0)
    {
        // sf_error.error("sph_harm", sf_error.ARG, "n should not be negative")
        return CUDART_NAN;
    }
    if (m < 0)
    {
        mp = -m;
        prefactor = poch(n + mp + 1, -2 * mp);
        if ((mp % 2) == 1)
        {
            prefactor = -prefactor;
        }
    }
    else
    {
        mp = m;
    }
    val = pmv_wrap(mp, n, x);
    if (m < 0)
    {
        val *= prefactor;
    }
    val *= sqrt((2*n + 1) / 4.0 / NPY_PI);
    val *= sqrt(poch(n + m + 1, -2 * m));

    complex<double> exponent(0, m * theta);
    val *= exp(exponent);

    return val;
}
"""
)


sph_harm = _core.create_ufunc(
    "cupyx_scipy_lpmv",
    ("iiff->F", "iidd->D", "llff->F", "lldd->D"),
    "out0 = out0_type(sph_harmonic(in0, in1, in2, in3));",
    preamble=sph_harmonic_definition,
    doc="""Spherical Harmonic.

    .. seealso:: :meth:`scipy.special.sph_harm`

    """,
)



import os

from cupy_backends.cuda.api cimport runtime


cdef list _elementwise_accelerators = []
cdef list _reduction_accelerators = []
cdef list _routine_accelerators = []


cdef int _get_accelerator(accelerator) except -1:
    if isinstance(accelerator, int):
        return accelerator
    if accelerator == 'cub':
        return ACCELERATOR_CUB
    if accelerator == 'cutensor':
        return ACCELERATOR_CUTENSOR
    if accelerator == 'cutensornet':
        return ACCELERATOR_CUTENSORNET
    raise ValueError('Unknown accelerator: {}'.format(accelerator))


def set_elementwise_accelerators(accelerators):
    global _elementwise_accelerators
    _elementwise_accelerators = [_get_accelerator(b) for b in accelerators]


def set_reduction_accelerators(accelerators):
    global _reduction_accelerators
    _reduction_accelerators = [_get_accelerator(b) for b in accelerators]


def set_routine_accelerators(accelerators):
    global _routine_accelerators
    _routine_accelerators = [_get_accelerator(b) for b in accelerators]


def get_elementwise_accelerators():
    return _elementwise_accelerators


def get_reduction_accelerators():
    return _reduction_accelerators


def get_routine_accelerators():
    return _routine_accelerators


cdef _set_default_accelerators():
    cdef str b, accelerator_names = os.getenv(
        'CUPY_ACCELERATORS', '' if runtime._is_hip_environment else 'cub')
    cdef list accelerators = [b for b in accelerator_names.split(',') if b]
    set_elementwise_accelerators(accelerators)
    set_reduction_accelerators(accelerators)
    set_routine_accelerators(accelerators)


_set_default_accelerators()



import distutils.ccompiler
import os
import os.path
import platform
import shutil
import sys
import subprocess
from typing import Any, Optional, List

from setuptools import Extension

from cupy_builder._context import Context
import cupy_builder.install_build as build


def _nvcc_gencode_options(cuda_version: int) -> List[str]:
    """Returns NVCC GPU code generation options."""

    if sys.argv == ['setup.py', 'develop']:
        return []

    envcfg = os.getenv('CUPY_NVCC_GENERATE_CODE', None)
    if envcfg is not None and envcfg != 'current':
        return ['--generate-code={}'.format(arch)
                for arch in envcfg.split(';') if len(arch) > 0]
    if envcfg == 'current' and build.get_compute_capabilities() is not None:
        ccs = build.get_compute_capabilities()
        arch_list = [
            f'compute_{cc}' if cc < 60 else (f'compute_{cc}', f'sm_{cc}')
            for cc in ccs]
    else:
        # The arch_list specifies virtual architectures, such as 'compute_61',
        # and real architectures, such as 'sm_61', for which the CUDA
        # input files are to be compiled.
        #
        # The syntax of an entry of the list is
        #
        #     entry ::= virtual_arch | (virtual_arch, real_arch)
        #
        # where virtual_arch is a string which means a virtual architecture and
        # real_arch is a string which means a real architecture.
        #
        # If a virtual architecture is supplied, NVCC generates a PTX code
        # the virtual architecture. If a pair of a virtual architecture and a
        # real architecture is supplied, NVCC generates a PTX code for the
        # virtual architecture as well as a cubin code for the real one.
        #
        # For example, making NVCC generate a PTX code for 'compute_60' virtual
        # architecture, the arch_list has an entry of 'compute_60'.
        #
        #     arch_list = ['compute_60']
        #
        # For another, making NVCC generate a PTX code for 'compute_61' virtual
        # architecture and a cubin code for 'sm_61' real architecture, the
        # arch_list has an entry of ('compute_61', 'sm_61').
        #
        #     arch_list = [('compute_61', 'sm_61')]
        #
        # See the documentation of each CUDA version for the list of supported
        # architectures:
        #
        #   https://docs.nvidia.com/cuda/cuda-compiler-driver-nvcc/index.html#options-for-steering-gpu-code-generation
        #
        # CuPy utilizes CUDA Minor Version Compatibility to support all CUDA
        # minor versions in a single binary package (e.g., `cupy-cuda12x`). To
        # achieve this, CUBIN must be generated for all supported compute
        # capabilities instead of PTX. This is because executing PTX requires
        # CUDA driver newer than the one used to compile the code, and we often
        # use the latest CUDA Driver to build our binary package. See also:
        #
        #   https://docs.nvidia.com/deploy/cuda-compatibility/index.html#application-considerations-for-minor-version-compatibility
        #
        # In addition, to allow running CuPy with future (not yet released)
        # GPUs, PTX for the latest architecture is also included as a
        # fallback. c.f.:
        #
        #   https://forums.developer.nvidia.com/t/software-migration-guide-for-nvidia-blackwell-rtx-gpus-a-guide-to-cuda-12-8-pytorch-tensorrt-and-llama-cpp/321330
        #
        # Jetson platforms are also targetted when built under aarch64. c.f.:
        #
        #   https://docs.nvidia.com/cuda/cuda-for-tegra-appnote/index.html#deployment-considerations-for-cuda-upgrade-package

        aarch64 = (platform.machine() == 'aarch64')
        if cuda_version >= 13000:
            arch_list = [('compute_75', 'sm_75'),
                         ('compute_80', 'sm_80'),
                         ('compute_86', 'sm_86'),
                         ('compute_89', 'sm_89'),
                         ('compute_90', 'sm_90'),
                         ('compute_100f', 'sm_100'),
                         ('compute_120f', 'sm_120'),
                         'compute_120']
            if aarch64:
                # JetPack
                arch_list += [
                    ('compute_87', 'sm_87'),    # Jetson (Orin)
                    ('compute_110', 'sm_110'),  # Jetson (Thor)
                ]
        elif cuda_version >= 12000:
            arch_list = [('compute_50', 'sm_50'),
                         ('compute_52', 'sm_52'),
                         ('compute_60', 'sm_60'),
                         ('compute_61', 'sm_61'),
                         ('compute_70', 'sm_70'),
                         ('compute_75', 'sm_75'),
                         ('compute_80', 'sm_80'),
                         ('compute_86', 'sm_86'),
                         ('compute_89', 'sm_89'),
                         ('compute_90', 'sm_90'),]
            if cuda_version < 12080:
                arch_list.append('compute_90')
            elif 12080 <= cuda_version < 12090:
                arch_list += [('compute_100', 'sm_100'),
                              ('compute_120', 'sm_120'),
                              'compute_100']
            elif 12090 <= cuda_version:
                arch_list += [('compute_100f', 'sm_100'),
                              ('compute_120f', 'sm_120'),
                              'compute_100']

            if aarch64:
                # JetPack 5 (CUDA 12.0-12.2) or JetPack 6 (CUDA 12.2+)
                arch_list += [
                    ('compute_72', 'sm_72'),  # Jetson (Xavier)
                    ('compute_87', 'sm_87'),  # Jetson (Orin)
                ]
        elif cuda_version >= 11080:
            arch_list = [('compute_35', 'sm_35'),
                         ('compute_37', 'sm_37'),
                         ('compute_50', 'sm_50'),
                         ('compute_52', 'sm_52'),
                         ('compute_60', 'sm_60'),
                         ('compute_61', 'sm_61'),
                         ('compute_70', 'sm_70'),
                         ('compute_75', 'sm_75'),
                         ('compute_80', 'sm_80'),
                         ('compute_86', 'sm_86'),
                         ('compute_89', 'sm_89'),
                         ('compute_90', 'sm_90'),
                         'compute_90']
            if aarch64:
                # JetPack 5 (CUDA 11.4/11.8)
                arch_list += [
                    ('compute_72', 'sm_72'),  # Jetson (Xavier)
                    ('compute_87', 'sm_87'),  # Jetson (Orin)
                ]
        elif cuda_version >= 11040:
            arch_list = [('compute_35', 'sm_35'),
                         ('compute_37', 'sm_37'),
                         ('compute_50', 'sm_50'),
                         ('compute_52', 'sm_52'),
                         ('compute_60', 'sm_60'),
                         ('compute_61', 'sm_61'),
                         ('compute_70', 'sm_70'),
                         ('compute_75', 'sm_75'),
                         ('compute_80', 'sm_80'),
                         ('compute_86', 'sm_86'),
                         'compute_86']
            if aarch64:
                # JetPack 5 (CUDA 11.4/11.8)
                arch_list += [
                    ('compute_72', 'sm_72'),  # Jetson (Xavier)
                    ('compute_87', 'sm_87'),  # Jetson (Orin)
                ]
        elif cuda_version >= 11020:
            arch_list = ['compute_35',
                         'compute_50',
                         ('compute_60', 'sm_60'),
                         ('compute_61', 'sm_61'),
                         ('compute_70', 'sm_70'),
                         ('compute_75', 'sm_75'),
                         ('compute_80', 'sm_80'),
                         ('compute_86', 'sm_86'),
                         'compute_86']
        else:
            # This should not happen.
            assert False

    options = []
    for arch in arch_list:
        if type(arch) is tuple:
            virtual_arch, real_arch = arch
            options.append('--generate-code=arch={},code={}'.format(
                virtual_arch, real_arch))
        else:
            options.append('--generate-code=arch={},code={}'.format(
                arch, arch))

    return options


class DeviceCompilerBase:
    """A class that invokes NVCC or HIPCC."""
    _context: Context

    def __init__(self, ctx: Context) -> None:
        self._context = ctx

    def _get_preprocess_options(self, ext: Extension) -> List[str]:
        # https://setuptools.pypa.io/en/latest/deprecated/distutils/apiref.html#distutils.core.Extension
        # https://github.com/pypa/setuptools/blob/v60.0.0/setuptools/_distutils/command/build_ext.py#L524-L526
        incdirs = ext.include_dirs[:]
        macros: List[Any] = ext.define_macros[:]
        for undef in ext.undef_macros:
            macros.append((undef,))
        return distutils.ccompiler.gen_preprocess_options(macros, incdirs)

    def spawn(self, commands: List[str]) -> None:
        print('Command:', commands)
        subprocess.check_call(commands)


class DeviceCompilerUnix(DeviceCompilerBase):

    def compile(self, obj: str, src: str, ext: Extension) -> None:
        if self._context.use_hip:
            self._compile_unix_hipcc(obj, src, ext)
        else:
            self._compile_unix_nvcc(obj, src, ext)

    def _compile_unix_nvcc(self, obj: str, src: str, ext: Extension) -> None:
        cc_args = self._get_preprocess_options(ext) + ['-c']

        # For CUDA C source files, compile them with NVCC.
        nvcc_path = build.get_nvcc_path()
        base_opts = build.get_compiler_base_options(nvcc_path)
        compiler_so = nvcc_path

        cuda_version = self._context.features['cuda'].get_version()
        postargs = _nvcc_gencode_options(cuda_version) + [
            '-Xfatbin=-compress-all', '-O2', '--compiler-options="-fPIC"',
            '--expt-relaxed-constexpr']
        num_threads = int(os.environ.get('CUPY_NUM_NVCC_THREADS', '2'))
        # Note: we only support CUDA 11.2+ since CuPy v13.0.0.
        # Bumping C++ standard from C++14 to C++17 for "if constexpr"
        postargs += ['--std=c++17',
                     f'-t{num_threads}',
                     '-Xcompiler=-fno-gnu-unique']
        print('NVCC options:', postargs)
        self.spawn(compiler_so + base_opts + cc_args + [src, '-o', obj] +
                   postargs)

    def _compile_unix_hipcc(self, obj: str, src: str, ext: Extension) -> None:
        cc_args = self._get_preprocess_options(ext) + ['-c']

        # For CUDA C source files, compile them with HIPCC.
        rocm_path = build.get_hipcc_path()
        base_opts = build.get_compiler_base_options(rocm_path)
        compiler_so = rocm_path

        postargs = ['-O2', '-fPIC', '--include', 'hip_runtime.h']
        # Note: we only support ROCm 4.3+ since CuPy v11.0.0.
        # Bumping C++ standard from C++14 to C++17 for "if constexpr"
        postargs += ['--std=c++17']
        print('HIPCC options:', postargs)
        self.spawn(compiler_so + base_opts + cc_args + [src, '-o', obj] +
                   postargs)


class DeviceCompilerWin32(DeviceCompilerBase):

    def compile(self, obj: str, src: str, ext: Extension) -> None:
        if self._context.use_hip:
            raise RuntimeError('ROCm is not supported on Windows')

        compiler_so = build.get_nvcc_path()
        cc_args = self._get_preprocess_options(ext) + ['-c']
        cuda_version = self._context.features['cuda'].get_version()
        postargs = _nvcc_gencode_options(cuda_version) + [
            '-Xfatbin=-compress-all', '-O2']
        # Note: we only support CUDA 11.2+ since CuPy v13.0.0.
        # MSVC 14.0 (2015) is deprecated for CUDA 11.2 but we need it
        # to build CuPy because some Python versions were built using it.
        # REF: https://wiki.python.org/moin/WindowsCompilers
        postargs += ['-allow-unsupported-compiler']
        # "/bigobj" to silence `fatal error C1128: number of sections exceeded
        # object file format limit`
        postargs += ['-Xcompiler', '/MD /bigobj', '-D_USE_MATH_DEFINES']
        # Bumping C++ standard from C++14 to C++17 for "if constexpr"
        num_threads = int(os.environ.get('CUPY_NUM_NVCC_THREADS', '2'))
        postargs += ['--std=c++17',
                     f'-t{num_threads}']
        cl_exe_path = self._find_host_compiler_path()
        if cl_exe_path is not None:
            print(f'Using host compiler at {cl_exe_path}')
            postargs += ['--compiler-bindir', cl_exe_path]
        print('NVCC options:', postargs)
        self.spawn(compiler_so + cc_args + [src, '-o', obj] + postargs)

    def _find_host_compiler_path(self) -> Optional[str]:
        # c.f. cupy.cuda.compiler._get_extra_path_for_msvc
        cl_exe = shutil.which('cl.exe')
        if cl_exe:
            # The compiler is already on PATH, no extra path needed.
            return None

        if self._context.win32_cl_exe_path is not None:
            return self._context.win32_cl_exe_path

        try:
            # See #8568, #8574, #8583.
            import setuptools.msvc
        except Exception:
            print('Warning: cl.exe could not be auto-detected; '
                  'setuptools.msvc could not be imported')
            return None

        vctools: List[str] = setuptools.msvc.EnvironmentInfo(
            platform.machine()).VCTools
        for path in vctools:
            cl_exe = os.path.join(path, 'cl.exe')
            if os.path.exists(cl_exe):
                return path
        print(f'Warning: cl.exe could not be found in {vctools}')
        return None



import cupy

cuvs_available = False
pylibraft_available = False
try:
    from cuvs.distance import pairwise_distance
    cuvs_available = True
except ImportError:
    try:
        # cuVS distance primitives were previously in pylibraft
        from pylibraft.distance import pairwise_distance
        pylibraft_available = True
    except ImportError:
        cuvs_available = False
        pylibraft_available = False


def _convert_to_type(X, out_type):
    return cupy.ascontiguousarray(X, dtype=out_type)


def _validate_pdist_input(X, m, n, metric_info, **kwargs):
    # get supported types
    types = metric_info.types
    # choose best type
    typ = types[types.index(X.dtype)] if X.dtype in types else types[0]
    # validate data
    X = _convert_to_type(X, out_type=typ)

    # validate kwargs
    _validate_kwargs = metric_info.validator
    if _validate_kwargs:
        kwargs = _validate_kwargs(X, m, n, **kwargs)
    return X, typ, kwargs


class MetricInfo:

    def __init__(self, canonical_name=None, aka=None,
                 validator=None, types=None):
        self.canonical_name_ = canonical_name
        self.aka_ = aka
        self.validator_ = validator
        self.types_ = types


_METRIC_INFOS = [
    MetricInfo(
        canonical_name="canberra",
        aka={'canberra'}
    ),
    MetricInfo(
        canonical_name="chebyshev",
        aka={"chebychev", "chebyshev", "cheby", "cheb", "ch"}
    ),
    MetricInfo(
        canonical_name="cityblock",
        aka={"cityblock", "cblock", "cb", "c"}
    ),
    MetricInfo(
        canonical_name="correlation",
        aka={"correlation", "co"}
    ),
    MetricInfo(
        canonical_name="cosine",
        aka={"cosine", "cos"}
    ),
    MetricInfo(
        canonical_name="hamming",
        aka={"matching", "hamming", "hamm", "ha", "h"},
        types=["double", "bool"]
    ),
    MetricInfo(
        canonical_name="euclidean",
        aka={"euclidean", "euclid", "eu", "e"},
    ),
    MetricInfo(
        canonical_name="jensenshannon",
        aka={"jensenshannon", "js"}
    ),
    MetricInfo(
        canonical_name="minkowski",
        aka={"minkowski", "mi", "m", "pnorm"}
    ),
    MetricInfo(
        canonical_name="russellrao",
        aka={"russellrao"},
        types=["bool"]
    ),
    MetricInfo(
        canonical_name="sqeuclidean",
        aka={"sqeuclidean", "sqe", "sqeuclid"}
    ),
    MetricInfo(
        canonical_name="hellinger",
        aka={"hellinger"}
    ),
    MetricInfo(
        canonical_name="kl_divergence",
        aka={"kl_divergence", "kl_div", "kld"}
    )


]

_METRICS = {info.canonical_name_: info for info in _METRIC_INFOS}
_METRIC_ALIAS = dict((alias, info)
                     for info in _METRIC_INFOS
                     for alias in info.aka_)

_METRICS_NAMES = list(_METRICS.keys())


def check_soft_dependencies():
    if not cuvs_available:
        if not pylibraft_available:
            raise RuntimeError('cuVS >= 24.12 or pylibraft < '
                               '24.12 should be installed to use this feature')


def minkowski(u, v, p):
    """Compute the Minkowski distance between two 1-D arrays.

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)
        p (float): The order of the norm of the difference
            :math:`{\\|u-v\\|}_p`. Note that for :math:`0 < p < 1`,
            the triangle inequality only holds with an additional
            multiplicative factor, i.e. it is only a quasi-metric.

    Returns:
        minkowski (double): The Minkowski distance between vectors `u` and `v`.
    """

    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "minkowski", p)

    return output_arr[0, 0]


def canberra(u, v):
    """Compute the Canberra distance between two 1-D arrays.

    The Canberra distance is defined as

    .. math::
        d(u, v) = \\sum_{i} \\frac{| u_i - v_i |}{|u_i| + |v_i|}

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        canberra (double): The Canberra distance between vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "canberra")

    return output_arr[0, 0]


def chebyshev(u, v):
    """Compute the Chebyshev distance between two 1-D arrays.

    The Chebyshev distance is defined as

    .. math::
        d(u, v) = \\max_{i} |u_i - v_i|

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        chebyshev (double): The Chebyshev distance between vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "chebyshev")

    return output_arr[0, 0]


def cityblock(u, v):
    """Compute the City Block (Manhattan) distance between two 1-D arrays.

    The City Block distance is defined as

    .. math::
        d(u, v) = \\sum_{i} |u_i - v_i|

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        cityblock (double): The City Block distance between
        vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "cityblock")

    return output_arr[0, 0]


def correlation(u, v):
    """Compute the correlation distance between two 1-D arrays.

    The correlation distance is defined as

    .. math::
        d(u, v) = 1 - \\frac{(u - \\bar{u}) \\cdot (v - \\bar{v})}{
        \\|(u - \\bar{u})\\|_2 \\|(v - \\bar{v})\\|_2}

    where :math:`\\bar{u}` is the mean of the elements of :math:`u` and
    :math:`x \\cdot y` is the dot product.

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        correlation (double): The correlation distance between
        vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "correlation")

    return output_arr[0, 0]


def cosine(u, v):
    """Compute the Cosine distance between two 1-D arrays.

    The Cosine distance is defined as

    .. math::
        d(u, v) = 1 - \\frac{u \\cdot v}{\\|u\\|_2 \\|v\\|_2}

    where :math:`x \\cdot y` is the dot product.

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        cosine (double): The Cosine distance between vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "cosine")

    return output_arr[0, 0]


def hamming(u, v):
    """Compute the Hamming distance between two 1-D arrays.

    The Hamming distance is defined as the proportion of elements
    in both `u` and `v` that are not in the exact same position:

    .. math::
        d(u, v) = \\frac{1}{n} \\sum_{k=0}^n u_i \\neq v_i

    where :math:`x \\neq y` is one if :math:`x` is different from :math:`y`
    and zero otherwise.

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        hamming (double): The Hamming distance between vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "hamming")

    return output_arr[0, 0]


def euclidean(u, v):
    """Compute the Euclidean distance between two 1-D arrays.

    The Euclidean distance is defined as

    .. math::
        d(u, v) = \\left(\\sum_{i} (u_i - v_i)^2\\right)^{\\sfrac{1}{2}}

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        euclidean (double): The Euclidean distance between vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "euclidean")

    return output_arr[0, 0]


def jensenshannon(u, v):
    """Compute the Jensen-Shannon distance between two 1-D arrays.

    The Jensen-Shannon distance is defined as

    .. math::
        d(u, v) = \\sqrt{\\frac{KL(u \\| m) + KL(v \\| m)}{2}}

    where :math:`KL` is the Kullback-Leibler divergence and :math:`m` is the
    pointwise mean of `u` and `v`.

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        jensenshannon (double): The Jensen-Shannon distance between
        vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "jensenshannon")

    return output_arr[0, 0]


def russellrao(u, v):
    """Compute the Russell-Rao distance between two 1-D arrays.

    The Russell-Rao distance is defined as the proportion of elements
    in both `u` and `v` that are in the exact same position:

    .. math::
        d(u, v) = \\frac{1}{n} \\sum_{k=0}^n u_i = v_i

    where :math:`x = y` is one if :math:`x` is different from :math:`y`
    and zero otherwise.

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        hamming (double): The Hamming distance between vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "russellrao")

    return output_arr[0, 0]


def sqeuclidean(u, v):
    """Compute the squared Euclidean distance between two 1-D arrays.

    The squared Euclidean distance is defined as

    .. math::
        d(u, v) = \\sum_{i} (u_i - v_i)^2

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        sqeuclidean (double): The squared Euclidean distance between
        vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "sqeuclidean")

    return output_arr[0, 0]


def hellinger(u, v):
    """Compute the Hellinger distance between two 1-D arrays.

    The Hellinger distance is defined as

    .. math::
        d(u, v) = \\frac{1}{\\sqrt{2}} \\sqrt{
            \\sum_{i} (\\sqrt{u_i} - \\sqrt{v_i})^2}

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        hellinger (double): The Hellinger distance between
        vectors `u` and `v`.
    """
    check_soft_dependencies()

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)

    pairwise_distance(u, v, output_arr, "hellinger")

    return output_arr[0, 0]


def kl_divergence(u, v):
    """Compute the Kullback-Leibler divergence between two 1-D arrays.

    The Kullback-Leibler divergence is defined as

    .. math::
        KL(U \\| V) = \\sum_{i} U_i \\log{\\left(\\frac{U_i}{V_i}\\right)}

    Args:
        u (array_like): Input array of size (N,)
        v (array_like): Input array of size (N,)

    Returns:
        kl_divergence (double): The Kullback-Leibler divergence between
        vectors `u` and `v`.
    """
    check_soft_dependencies()

    u = cupy.asarray(u)
    v = cupy.asarray(v)

    u_order = "F" if cupy.isfortran(u) else "C"
    v_order = "F" if cupy.isfortran(v) else "C"

    if u_order != v_order:
        raise ValueError('u and v must have the same layout '
                         '(u.order=%s, v.order=%s' % (u_order, v_order))

    output_arr = cupy.empty((1, 1), dtype=u.dtype, order=u_order)
    pairwise_distance(u, v, output_arr, "kl_divergence")

    return output_arr[0, 0]


def cdist(XA, XB, metric='euclidean', out=None, **kwargs):
    """Compute distance between each pair of the two collections of inputs.

    Args:
        XA (array_like): An :math:`m_A` by :math:`n` array of :math:`m_A`
            original observations in an :math:`n`-dimensional space.
            Inputs are converted to float type.
        XB (array_like): An :math:`m_B` by :math:`n` array of :math:`m_B`
            original observations in an :math:`n`-dimensional space.
            Inputs are converted to float type.
        metric (str, optional): The distance metric to use.
            The distance function can be 'canberra', 'chebyshev',
            'cityblock', 'correlation', 'cosine', 'euclidean', 'hamming',
            'hellinger', 'jensenshannon', 'kl_divergence', 'matching',
            'minkowski', 'russellrao', 'sqeuclidean'.
        out (cupy.ndarray, optional): The output array. If not None, the
            distance matrix Y is stored in this array.
        **kwargs (dict, optional): Extra arguments to `metric`: refer to each
            metric documentation for a list of all possible arguments.
            Some possible arguments:
            p (float): The p-norm to apply for Minkowski, weighted and
            unweighted. Default: 2.0

    Returns:
        Y (cupy.ndarray): A :math:`m_A` by :math:`m_B` distance matrix is
            returned. For each :math:`i` and :math:`j`, the metric
            ``dist(u=XA[i], v=XB[j])`` is computed and stored in the
            :math:`ij` th entry.
    """
    check_soft_dependencies()

    if pylibraft_available or \
            (cuvs_available and XA.dtype not in ['float32', 'float64']):
        XA = cupy.asarray(XA, dtype='float32')

    if pylibraft_available or \
            (cuvs_available and XB.dtype not in ['float32', 'float64']):
        XB = cupy.asarray(XB, dtype='float32')

    XA_order = "F" if cupy.isfortran(XA) else "C"
    XB_order = "F" if cupy.isfortran(XB) else "C"

    if XA_order != XB_order:
        raise ValueError('XA and XB must have the same layout '
                         '(XA.order=%s, XB.order=%s' % (XA_order, XB_order))

    s = XA.shape
    sB = XB.shape

    if len(s) != 2:
        raise ValueError('XA must be a 2-dimensional array.')
    if len(sB) != 2:
        raise ValueError('XB must be a 2-dimensional array.')
    if s[1] != sB[1]:
        raise ValueError('XA and XB must have the same number of columns '
                         '(i.e. feature dimension.)')

    mA = s[0]
    mB = sB[0]

    p = kwargs["p"] if "p" in kwargs else 2.0

    if out is not None:
        if (pylibraft_available and out.dtype != 'float32') or \
                (cuvs_available and out.dtype not in ['float32', 'float64']):
            out_order = "F" if cupy.isfortran(out) else "C"
            if out_order != XA_order:
                raise ValueError('out must have same layout as input '
                                 '(out.order=%s)' % out_order)
            out = out.astype('float32', copy=False)
        if out.shape != (mA, mB):
            cupy.resize(out, (mA, mB))

    if isinstance(metric, str):
        mstr = metric.lower()
        metric_info = _METRIC_ALIAS.get(mstr, None)
        if metric_info is not None:
            output_arr = out if out is not None else cupy.empty((mA, mB),
                                                                dtype=XA.dtype,
                                                                order=XA_order)
            pairwise_distance(XA, XB, output_arr, metric, p)
            return output_arr
        else:
            raise ValueError('Unknown Distance Metric: %s' % mstr)
    else:
        raise TypeError('2nd argument metric must be a string identifier')


def pdist(X, metric='euclidean', *, out=None, **kwargs):
    """Compute distance between observations in n-dimensional space.

    Args:
        X (array_like): An :math:`m` by :math:`n` array of :math:`m`
            original observations in an :math:`n`-dimensional space.
            Inputs are converted to float type.
        metric (str, optional): The distance metric to use.
            The distance function can be 'canberra', 'chebyshev',
            'cityblock', 'correlation', 'cosine', 'euclidean', 'hamming',
            'hellinger', 'jensenshannon', 'kl_divergence', 'matching',
            'minkowski', 'russellrao', 'sqeuclidean'.
        out (cupy.ndarray, optional): The output array. If not None, the
            distance matrix Y is stored in this array.
        **kwargs (dict, optional): Extra arguments to `metric`: refer to each
            metric documentation for a list of all possible arguments.
            Some possible arguments:
            p (float): The p-norm to apply for Minkowski, weighted and
            unweighted. Default: 2.0

    Returns:
        Y (cupy.ndarray):
            Returns a condensed distance matrix Y. For each :math:`i` and
            :math:`j` and (where :math:`i < j < m`), where m is the number of
            original observations. The metric ``dist(u=X[i], v=X[j])`` is
            computed and stored in entry
            ``m * i + j - ((i + 2) * (i + 1)) // 2``.
    """
    all_dist = cdist(X, X, metric=metric, out=out, **kwargs)
    up_idx = cupy.triu_indices_from(all_dist, 1)
    return all_dist[up_idx]


def distance_matrix(x, y, p=2.0):
    """Compute the distance matrix.

    Returns the matrix of all pair-wise distances.

    Args:
        x (array_like): Matrix of M vectors in K dimensions.
        y (array_like): Matrix of N vectors in K dimensions.
        p (float): Which Minkowski p-norm to use (1 <= p <= infinity).
            Default=2.0
    Returns:
        result (cupy.ndarray): Matrix containing the distance from every
            vector in `x` to every vector in `y`, (size M, N).
    """
    x = cupy.asarray(x)
    m, k = x.shape
    y = cupy.asarray(y)
    n, kk = y.shape

    if k != kk:
        raise ValueError("x contains %d-dimensional vectors but y "
                         "contains %d-dimensional vectors" % (k, kk))

    return cdist(x, y, metric="minkowski", p=p)



import unittest

import cupy
from cupy import cuda
from cupy import testing
import numpy
from numpy import testing as np_testing


class TestArrayGet(unittest.TestCase):

    def setUp(self):
        self.stream = cuda.Stream.null

    def check_get(self, f, stream, order='C', blocking=True):
        a_gpu = f(cupy)
        a_cpu = a_gpu.get(stream, order=order, blocking=blocking)
        if stream:
            stream.synchronize()
        b_cpu = f(numpy)
        np_testing.assert_array_equal(a_cpu, b_cpu)
        if order == 'F' or (order == 'A' and a_gpu.flags.f_contiguous):
            assert a_cpu.flags.f_contiguous
        else:
            assert a_cpu.flags.c_contiguous

    @testing.for_orders('CFA')
    @testing.for_all_dtypes()
    def test_contiguous_array(self, dtype, order):
        def contiguous_array(xp):
            return testing.shaped_arange((3,), xp, dtype, order)
        self.check_get(contiguous_array, None, order)

    @testing.for_orders('CFA')
    @testing.for_all_dtypes()
    def test_non_contiguous_array(self, dtype, order):
        def non_contiguous_array(xp):
            return testing.shaped_arange((3, 3), xp, dtype, order)[0::2, 0::2]
        self.check_get(non_contiguous_array, None, order)

    @testing.for_orders('CFA')
    @testing.for_all_dtypes()
    def test_contiguous_array_stream(self, dtype, order):
        def contiguous_array(xp):
            return testing.shaped_arange((3,), xp, dtype, order)
        self.check_get(contiguous_array, self.stream, order)

    @testing.for_orders('CFA')
    @testing.for_all_dtypes()
    def test_contiguous_array_stream_nonblocking(self, dtype, order):
        # Note: This is just a smoking gun test, the real test is done for
        # testing cupy.asnumpy(), which under the hood calls .get().
        def contiguous_array(xp):
            return testing.shaped_arange((3,), xp, dtype, order)
        self.check_get(contiguous_array, self.stream, order, False)

    @testing.for_orders('CFA')
    @testing.for_all_dtypes()
    def test_non_contiguous_array_stream(self, dtype, order):
        def non_contiguous_array(xp):
            return testing.shaped_arange((3, 3), xp, dtype, order)[0::2, 0::2]
        self.check_get(non_contiguous_array, self.stream)

    @testing.multi_gpu(2)
    @testing.for_orders('CFA')
    @testing.for_all_dtypes()
    def test_get_multigpu(self, dtype, order):
        with cuda.Device(1):
            src = testing.shaped_arange((2, 3), cupy, dtype, order)
            src = cupy.asfortranarray(src)
        with cuda.Device(0):
            dst = src.get()
        expected = testing.shaped_arange((2, 3), numpy, dtype, order)
        np_testing.assert_array_equal(dst, expected)


class TestArrayGetWithOut(unittest.TestCase):

    def setUp(self):
        self.stream = cuda.Stream.null

    def check_get(self, f, out, stream):
        a_gpu = f(cupy)
        a_cpu = a_gpu.get(stream, out=out)
        if stream:
            stream.synchronize()
        b_cpu = f(numpy)
        assert a_cpu is out
        np_testing.assert_array_equal(a_cpu, b_cpu)

    @testing.for_orders('CF')
    @testing.for_all_dtypes()
    def test_contiguous_array(self, dtype, order):
        def contiguous_array(xp):
            return testing.shaped_arange((3,), xp, dtype, order)
        out = numpy.empty((3,), dtype, order)
        self.check_get(contiguous_array, out, None)

    @testing.for_orders('CF')
    @testing.for_all_dtypes()
    def test_contiguous_array_cross(self, dtype, order):
        def contiguous_array(xp):
            return testing.shaped_arange((3,), xp, dtype, order)
        out_order = 'C' if order == 'F' else 'F'
        out = numpy.empty((3,), dtype, out_order)
        self.check_get(contiguous_array, out, None)

    @testing.for_orders('CF')
    @testing.for_all_dtypes()
    def test_contiguous_array_with_error(self, dtype, order):
        out = numpy.empty((3, 3), dtype)[0:2, 0:2]
        with self.assertRaises(RuntimeError):
            a_gpu = testing.shaped_arange((3, 3), cupy, dtype, order)[0:2, 0:2]
            a_gpu.get(out=out)

    @testing.for_orders('CF')
    @testing.for_all_dtypes()
    def test_non_contiguous_array(self, dtype, order):
        def non_contiguous_array(xp):
            return testing.shaped_arange((3, 3), xp, dtype, order)[0::2, 0::2]
        out = numpy.empty((2, 2), dtype, order)
        self.check_get(non_contiguous_array, out, None)

    @testing.for_orders('CF')
    @testing.for_all_dtypes()
    def test_contiguous_array_stream(self, dtype, order):
        def contiguous_array(xp):
            return testing.shaped_arange((3,), xp, dtype, order)
        out = numpy.empty((3,), dtype, order)
        self.check_get(contiguous_array, out, self.stream)

    @testing.for_orders('CF')
    @testing.for_all_dtypes()
    def test_non_contiguous_array_stream(self, dtype, order):
        def non_contiguous_array(xp):
            return testing.shaped_arange((3, 3), xp, dtype, order)[0::2, 0::2]
        out = numpy.empty((2, 2), dtype, order)
        self.check_get(non_contiguous_array, out, self.stream)

    @testing.multi_gpu(2)
    @testing.for_orders('CF')
    @testing.for_all_dtypes()
    def test_get_multigpu(self, dtype, order):
        with cuda.Device(1):
            src = testing.shaped_arange((2, 3), cupy, dtype, order)
            src = cupy.asfortranarray(src)
        with cuda.Device(0):
            dst = numpy.empty((2, 3), dtype, order)
            src.get(out=dst)
        expected = testing.shaped_arange((2, 3), numpy, dtype, order)
        np_testing.assert_array_equal(dst, expected)



from collections.abc import Sequence

from typing import Any, Literal, SupportsIndex, TypeVar, Union

import numpy
from cupy._core import core

from numpy.typing import ArrayLike  # NOQA
from numpy.typing import DTypeLike  # NOQA
from numpy.typing import NBitBase  # NOQA


# Shapes
_Shape = tuple[int, ...]
_ShapeLike = Union[SupportsIndex, Sequence[SupportsIndex]]

_OrderKACF = Literal[None, "K", "A", "C", "F"]
_OrderCF = Literal[None, "C", "F"]
_ScalarType_co = TypeVar("_ScalarType_co", bound=numpy.generic, covariant=True)
NDArray = core.ndarray[Any, numpy.dtype[_ScalarType_co]]



import math

import cupy


def float_factorial(n):
    """Compute the factorial and return as a float

    Returns infinity when result is too large for a double
    """
    return float(math.factorial(n)) if n < 171 else cupy.inf


def _asarray_validated(a, check_finite=True,
                       sparse_ok=False, objects_ok=False, mask_ok=False,
                       as_inexact=False):
    """Helper function for SciPy argument validation.

    Many CuPy linear algebra functions do support arbitrary array-like
    input arguments. Examples of commonly unsupported inputs include
    matrices containing inf/nan, sparse matrix representations, and
    matrices with complicated elements.

    Parameters
    ----------
    a : array-like
        The array-like input
    check_finite : bool, optional
        By default True. To check whether the input matrices contain
        only finite numbers. Disabling may give a performance gain,
        but may result in problems (crashes, non-termination) if the
        inputs do contain infinites or NaNs
    sparse_ok : bool, optional
        By default False. True if cupy sparse matrices are allowed
    objects_ok : bool, optional
        By default False. True if arrays with dype('O') are allowed
    mask_ok : bool, optional
        By default False. True if masked arrays are allowed.
    as_inexact : bool, optional
        By default False. True to convert the input array to a
        cupy.inexact dtype

    Returns
    -------
    ret : cupy.ndarray
        The converted validated array

    """

    if not sparse_ok:
        import cupyx.scipy.sparse
        if cupyx.scipy.sparse.issparse(a):
            msg = ('Sparse matrices are not supported by this function. '
                   'Perhaps one of the cupyx.scipy.sparse.linalg functions '
                   'would work instead.')
            raise ValueError(msg)

    # TODO: remove these comments when CuPy supports masked arrays
    # Ref Issue: https://github.com/cupy/cupy/issues/2225
    # if not mask_ok:
    #     if cupy.ma.isMaskedArray(a):
    #         raise ValueError('masked arrays are not supported')

    # TODO: remove these comments when CuPy supports 'object' dtype
    # if not objects_ok:
    #    if a.dtype is cupy.dtype('O'):
    #        raise ValueError('object arrays are not supported')

    if as_inexact:
        if not cupy.issubdtype(a, cupy.inexact):
            a = a.astype(dtype=cupy.float64)

    return a



import numpy
import pytest
try:
    import scipy
except ImportError:
    pass

import cupy
from cupy import testing
from cupyx import signal


def _numpy_pulse_preprocess(x, normalize, window):
    if window is not None:
        n = x.shape[-1]
        if callable(window):
            x *= window(numpy.fft.fftfreq(n))
        else:
            from scipy.signal import get_window
            x *= get_window(window, n, False)

    if normalize:
        x /= numpy.linalg.norm(x)

    return x


def _numpy_pulse_doppler(x, window):
    xT = _numpy_pulse_preprocess(x.T, False, window)
    fft_x = numpy.fft.fft(xT, xT.shape[1]).T
    return fft_x.astype(x.dtype.char.upper())


def _numpy_pulse_compression(x, t, normalize, window):
    n = x.shape[1]
    dtype = numpy.result_type(x, t)
    t = _numpy_pulse_preprocess(t, normalize, window)
    fft_x = numpy.fft.fft(x, n)
    fft_t = numpy.fft.fft(t, n)
    out = numpy.fft.ifft(fft_x * fft_t.conj(), n)
    if dtype.kind != 'c':
        out = out.real
    return out.astype(dtype)


tol = {
    numpy.float32: 2e-3,
    numpy.float64: 1e-7,
    numpy.complex64: 2e-3,
    numpy.complex128: 1e-7,
}


@pytest.mark.parametrize('normalize', [True, False])
@pytest.mark.parametrize('window', [None, 'hamming', numpy.negative])
@testing.for_dtypes('fdFD')
@testing.numpy_cupy_allclose(rtol=tol, contiguous_check=False)
@testing.with_requires("scipy")
def test_pulse_compression(xp, normalize, window, dtype):
    if (
            numpy.dtype(dtype).char in "fF" and
            numpy.__version__ >= numpy.lib.NumpyVersion('2.0.0')
    ):
        return xp.array([])  # Skip

    x = testing.shaped_random((8, 700), xp=xp, dtype=dtype)
    template = testing.shaped_random((100,), xp=xp, dtype=dtype)

    if xp is cupy:
        return signal.pulse_compression(x, template, normalize, window)
    else:
        assert xp is numpy
        # May use scipy inside _numpy_pulse_compression
        return _numpy_pulse_compression(x, template, normalize, window)


@pytest.mark.parametrize('window', [None, 'hamming', numpy.negative])
@testing.for_dtypes('fdFD')
@testing.numpy_cupy_allclose(rtol=tol, contiguous_check=False)
@testing.with_requires("scipy")
def test_pulse_doppler(xp, window, dtype):
    x = testing.shaped_random((8, 700), xp=xp, dtype=dtype)

    if xp is cupy:
        return signal.pulse_doppler(x, window)
    else:
        assert xp is numpy
        # May use scipy inside _numpy_pulse_doppler
        return _numpy_pulse_doppler(x, window)


@testing.for_all_dtypes()
def test_cfar_alpha(dtype):
    N = 128
    pfa = testing.shaped_random((128, 128), xp=cupy, dtype=dtype)
    gpu = signal.cfar_alpha(pfa, N)
    cpu = N * (pfa ** (-1.0 / N) - 1)
    cupy.testing.assert_allclose(gpu, cpu)


@pytest.mark.parametrize(
    "size,gc,rc", [(100, 1, 5), (11, 2, 3), (100, 10, 20)])
@testing.for_float_dtypes(no_float16=True)
@testing.numpy_cupy_allclose(rtol=2e-6, type_check=False)
@testing.with_requires("scipy")
def test_ca_cfar1d(xp, dtype, size, gc, rc):
    array = testing.shaped_random((size,), xp=xp, dtype=dtype)
    if xp is cupy:
        return signal.ca_cfar(array, gc, rc)
    else:
        assert xp is numpy
        weight = numpy.ones(((rc + gc) * 2 + 1,), dtype=dtype)
        weight[rc:-rc] = 0
        alpha = numpy.zeros((size,), dtype=dtype)
        alpha[gc+rc:-gc-rc] = signal.cfar_alpha(1e-3, 2 * rc)
        out = scipy.ndimage.convolve1d(array, weight) * alpha / (2 * rc)
        return out, array - out > 0


@pytest.mark.parametrize(
    "size,gc,rc", [(1, 1, 1), (10, 2, 3), (10, 0, 5), (10, 5, 0)])
def test_ca_cfar1d_failures(size, gc, rc):
    with pytest.raises(ValueError):
        _, _ = signal.ca_cfar(cupy.zeros(size), gc, rc)


@pytest.mark.parametrize(
    "shape,gc,rc", [((10, 10), (1, 1), (2, 2)), ((10, 100), (1, 10), (2, 20))])
@testing.for_float_dtypes(no_float16=True)
@testing.numpy_cupy_allclose(rtol=2e-6, type_check=False)
@testing.with_requires("scipy")
def test_ca_cfar2d(xp, dtype, shape, gc, rc):
    array = testing.shaped_random(shape, xp=xp, dtype=dtype)
    if xp is cupy:
        return signal.ca_cfar(array, gc, rc)
    else:
        assert xp is numpy
        rcx, rcy = rc
        gcx, gcy = gc
        weight = numpy.ones(
            ((rcx + gcx) * 2 + 1, (rcy + gcy) * 2 + 1), dtype=dtype)
        weight[rcx:-rcx, rcy:-rcy] = 0
        alpha = numpy.zeros(shape, dtype=dtype)
        N = weight.size - (2 * gcx + 1) * (2 * gcy + 1)
        alpha[gcx+rcx:-gcx-rcx, gcy+rcy:-gcy-rcy] = signal.cfar_alpha(1e-3, N)
        out = scipy.ndimage.convolve(array, weight) * alpha / N
        return out, array - out > 0


@pytest.mark.parametrize(
    "shape,gc,rc", [((3, 3), (1, 2), (1, 10)),
                    ((3, 3), (1, 1), (10, 1)),
                    ((5, 5), (3, 3), (3, 3))])
def test_ca_cfar2d_failures(shape, gc, rc):
    with pytest.raises(ValueError):
        _, _ = signal.ca_cfar(cupy.zeros(shape), gc, rc)


@testing.for_float_dtypes(no_float16=True)
def test_mvdr(dtype):
    x = cupy.array([[1, 2, 3], [4, 5, 7]], dtype=dtype)
    sv = cupy.array([1, 2], dtype=dtype)
    out = signal.mvdr(x, sv)
    assert out.dtype == dtype
    testing.assert_allclose(out, [-2, 1.5])



from libc.stdint cimport int32_t, uint32_t, int64_t, uint64_t, intptr_t

###############################################################################
# Enum
###############################################################################

cpdef enum:
    # cusparseLtSparsity_t
    CUSPARSELT_SPARSITY_50_PERCENT = 0

    # cusparseLtMatDescAttribute_t
    CUSPARSELT_MAT_NUM_BATCHES = 0   # READ/WRITE
    CUSPARSELT_MAT_BATCH_STRIDE = 1  # READ/WRITE

    # cusparseComputeType
    CUSPARSE_COMPUTE_16F = 0
    CUSPARSE_COMPUTE_32I = 1
    CUSPARSE_COMPUTE_TF32 = 2
    CUSPARSE_COMPUTE_TF32_FAST = 3

    # cusparseLtMatmulDescAttribute_t
    CUSPARSELT_MATMUL_ACTIVATION_RELU = 0             # READ/WRITE
    CUSPARSELT_MATMUL_ACTIVATION_RELU_UPPERBOUND = 1  # READ/WRITE
    CUSPARSELT_MATMUL_ACTIVATION_RELU_THRESHOLD = 2   # READ/WRITE
    CUSPARSELT_MATMUL_ACTIVATION_GELU = 3             # READ/WRITE
    CUSPARSELT_MATMUL_BIAS_STRIDE = 4                 # READ/WRITE
    CUSPARSELT_MATMUL_BIAS_POINTER = 5                # READ/WRITE

    # cusparseLtMatmulAlg_t
    CUSPARSELT_MATMUL_ALG_DEFAULT = 0

    # cusparseLtMatmulAlgAttribute_t
    CUSPARSELT_MATMUL_ALG_CONFIG_ID = 0      # NOQA, READ/WRITE
    CUSPARSELT_MATMUL_ALG_CONFIG_MAX_ID = 1  # NOQA, READ-ONLY
    CUSPARSELT_MATMUL_SEARCH_ITERATIONS = 2  # NOQA, READ/WRITE

    # cusparseLtPruneAlg_t
    CUSPARSELT_PRUNE_SPMMA_TILE = 0
    CUSPARSELT_PRUNE_SPMMA_STRIP = 1



from cupyx.scipy.ndimage._filters import correlate  # NOQA
from cupyx.scipy.ndimage._filters import convolve  # NOQA
from cupyx.scipy.ndimage._filters import correlate1d  # NOQA
from cupyx.scipy.ndimage._filters import convolve1d  # NOQA
from cupyx.scipy.ndimage._filters import uniform_filter1d  # NOQA
from cupyx.scipy.ndimage._filters import uniform_filter  # NOQA
from cupyx.scipy.ndimage._filters import gaussian_filter1d  # NOQA
from cupyx.scipy.ndimage._filters import gaussian_filter  # NOQA
from cupyx.scipy.ndimage._filters import prewitt  # NOQA
from cupyx.scipy.ndimage._filters import sobel  # NOQA
from cupyx.scipy.ndimage._filters import generic_laplace  # NOQA
from cupyx.scipy.ndimage._filters import laplace  # NOQA
from cupyx.scipy.ndimage._filters import gaussian_laplace  # NOQA
from cupyx.scipy.ndimage._filters import generic_gradient_magnitude  # NOQA
from cupyx.scipy.ndimage._filters import gaussian_gradient_magnitude  # NOQA
from cupyx.scipy.ndimage._filters import minimum_filter  # NOQA
from cupyx.scipy.ndimage._filters import maximum_filter  # NOQA
from cupyx.scipy.ndimage._filters import minimum_filter1d  # NOQA
from cupyx.scipy.ndimage._filters import maximum_filter1d  # NOQA
from cupyx.scipy.ndimage._filters import median_filter  # NOQA
from cupyx.scipy.ndimage._filters import rank_filter  # NOQA
from cupyx.scipy.ndimage._filters import percentile_filter  # NOQA
from cupyx.scipy.ndimage._filters import generic_filter  # NOQA
from cupyx.scipy.ndimage._filters import generic_filter1d  # NOQA

from cupyx.scipy.ndimage._fourier import fourier_ellipsoid  # NOQA
from cupyx.scipy.ndimage._fourier import fourier_gaussian  # NOQA
from cupyx.scipy.ndimage._fourier import fourier_shift  # NOQA
from cupyx.scipy.ndimage._fourier import fourier_uniform  # NOQA

from cupyx.scipy.ndimage._interpolation import affine_transform  # NOQA
from cupyx.scipy.ndimage._interpolation import map_coordinates  # NOQA
from cupyx.scipy.ndimage._interpolation import rotate  # NOQA
from cupyx.scipy.ndimage._interpolation import shift  # NOQA
from cupyx.scipy.ndimage._interpolation import spline_filter  # NOQA
from cupyx.scipy.ndimage._interpolation import spline_filter1d  # NOQA
from cupyx.scipy.ndimage._interpolation import zoom  # NOQA

from cupyx.scipy.ndimage._measurements import label  # NOQA
from cupyx.scipy.ndimage._measurements import sum  # NOQA
from cupyx.scipy.ndimage._measurements import sum_labels  # NOQA
from cupyx.scipy.ndimage._measurements import mean  # NOQA
from cupyx.scipy.ndimage._measurements import variance  # NOQA
from cupyx.scipy.ndimage._measurements import standard_deviation  # NOQA
from cupyx.scipy.ndimage._measurements import minimum  # NOQA
from cupyx.scipy.ndimage._measurements import maximum  # NOQA
from cupyx.scipy.ndimage._measurements import minimum_position  # NOQA
from cupyx.scipy.ndimage._measurements import maximum_position  # NOQA
from cupyx.scipy.ndimage._measurements import median  # NOQA
from cupyx.scipy.ndimage._measurements import extrema  # NOQA
from cupyx.scipy.ndimage._measurements import center_of_mass  # NOQA
from cupyx.scipy.ndimage._measurements import histogram  # NOQA
from cupyx.scipy.ndimage._measurements import labeled_comprehension  # NOQA
from cupyx.scipy.ndimage._measurements import value_indices  # NOQA

from cupyx.scipy.ndimage._morphology import generate_binary_structure  # NOQA
from cupyx.scipy.ndimage._morphology import iterate_structure  # NOQA
from cupyx.scipy.ndimage._morphology import binary_erosion  # NOQA
from cupyx.scipy.ndimage._morphology import binary_dilation  # NOQA
from cupyx.scipy.ndimage._morphology import binary_opening  # NOQA
from cupyx.scipy.ndimage._morphology import binary_closing  # NOQA
from cupyx.scipy.ndimage._morphology import binary_hit_or_miss  # NOQA
from cupyx.scipy.ndimage._morphology import binary_fill_holes  # NOQA
from cupyx.scipy.ndimage._morphology import binary_propagation  # NOQA
from cupyx.scipy.ndimage._morphology import grey_erosion  # NOQA
from cupyx.scipy.ndimage._morphology import grey_dilation  # NOQA
from cupyx.scipy.ndimage._morphology import grey_closing  # NOQA
from cupyx.scipy.ndimage._morphology import grey_opening  # NOQA
from cupyx.scipy.ndimage._morphology import morphological_gradient  # NOQA
from cupyx.scipy.ndimage._morphology import morphological_laplace  # NOQA
from cupyx.scipy.ndimage._morphology import white_tophat  # NOQA
from cupyx.scipy.ndimage._morphology import black_tophat  # NOQA
from cupyx.scipy.ndimage._distance_transform import distance_transform_edt  # NOQA



import unittest
import warnings

import numpy
import cupy
from cupy import testing
import cupyx.scipy.linalg
if cupyx.scipy._scipy_available:
    import scipy.linalg


# TODO: After the feature is released
# requires_scipy_linalg_backend = testing.with_requires('scipy>=1.x.x')
requires_scipy_linalg_backend = unittest.skip(
    'scipy.linalg backend feature has not been released'
)


@testing.parameterize(*testing.product({
    'shape': [(1, 1), (2, 2), (3, 3), (5, 5), (1, 5), (5, 1), (2, 5), (5, 2)],
}))
@testing.fix_random()
@testing.with_requires('scipy')
class TestLUFactor(unittest.TestCase):

    @testing.for_dtypes('fdFD')
    def test_lu_factor(self, dtype):
        if self.shape[0] != self.shape[1]:
            self.skipTest(
                'skip non-square tests since scipy.lu_factor requires square')
        a_cpu = testing.shaped_random(self.shape, numpy, dtype=dtype)
        a_gpu = cupy.asarray(a_cpu)
        result_cpu = scipy.linalg.lu_factor(a_cpu)
        result_gpu = cupyx.scipy.linalg.lu_factor(a_gpu)
        assert len(result_cpu) == len(result_gpu)
        assert result_cpu[0].dtype == result_gpu[0].dtype
        assert result_cpu[1].dtype == result_gpu[1].dtype
        cupy.testing.assert_allclose(result_cpu[0], result_gpu[0], atol=1e-5)
        cupy.testing.assert_array_equal(result_cpu[1], result_gpu[1])

    def check_lu_factor_reconstruction(self, A):
        m, n = self.shape
        lu, piv = cupyx.scipy.linalg.lu_factor(A)
        # extract ``L`` and ``U`` from ``lu``
        L = cupy.tril(lu, k=-1)
        cupy.fill_diagonal(L, 1.)
        L = L[:, :m]
        U = cupy.triu(lu)
        U = U[:n, :]
        # check output shapes
        assert lu.shape == (m, n)
        assert L.shape == (m, min(m, n))
        assert U.shape == (min(m, n), n)
        assert piv.shape == (min(m, n),)
        # apply pivot (on CPU since slaswp is not available in cupy)
        piv = cupy.asnumpy(piv)
        rows = numpy.arange(m)
        for i, row in enumerate(piv):
            if i != row:
                rows[i], rows[row] = rows[row], rows[i]
        PA = A[rows]
        # check that reconstruction is close to original
        LU = L.dot(U)
        cupy.testing.assert_allclose(LU, PA, atol=1e-5)

    @testing.for_dtypes('fdFD')
    def test_lu_factor_reconstruction(self, dtype):
        A = testing.shaped_random(self.shape, cupy, dtype=dtype)
        self.check_lu_factor_reconstruction(A)

    @testing.for_dtypes('fdFD')
    def test_lu_factor_reconstruction_singular(self, dtype):
        if self.shape[0] != self.shape[1]:
            self.skipTest(
                'skip non-square tests since scipy.lu_factor requires square')
        A = testing.shaped_random(self.shape, cupy, dtype=dtype)
        A -= A.mean(axis=0, keepdims=True)
        A -= A.mean(axis=1, keepdims=True)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            self.check_lu_factor_reconstruction(A)


@testing.parameterize(*testing.product({
    'shape': [(1, 1), (2, 2), (3, 3), (5, 5), (1, 5), (5, 1), (2, 5), (5, 2)],
    'permute_l': [False, True],
}))
@testing.fix_random()
@testing.with_requires('scipy')
class TestLU(unittest.TestCase):

    @testing.for_dtypes('fdFD')
    def test_lu(self, dtype):
        a_cpu = testing.shaped_random(self.shape, numpy, dtype=dtype)
        a_gpu = cupy.asarray(a_cpu)
        result_cpu = scipy.linalg.lu(a_cpu, permute_l=self.permute_l)
        result_gpu = cupyx.scipy.linalg.lu(a_gpu, permute_l=self.permute_l)
        assert len(result_cpu) == len(result_gpu)
        if not self.permute_l:
            # check permutation matrix
            result_cpu = list(result_cpu)
            result_gpu = list(result_gpu)
            P_cpu = result_cpu.pop(0)
            P_gpu = result_gpu.pop(0)
            cupy.testing.assert_array_equal(P_gpu, P_cpu)
        cupy.testing.assert_allclose(result_gpu[0], result_cpu[0], atol=1e-5)
        cupy.testing.assert_allclose(result_gpu[1], result_cpu[1], atol=1e-5)

    @testing.for_dtypes('fdFD')
    def test_lu_reconstruction(self, dtype):
        m, n = self.shape
        A = testing.shaped_random(self.shape, cupy, dtype=dtype)
        if self.permute_l:
            PL, U = cupyx.scipy.linalg.lu(A, permute_l=self.permute_l)
            PLU = PL @ U
        else:
            P, L, U = cupyx.scipy.linalg.lu(A, permute_l=self.permute_l)
            PLU = P @ L @ U
        # check that reconstruction is close to original
        cupy.testing.assert_allclose(PLU, A, atol=1e-5)


@testing.parameterize(*testing.product({
    'trans': [0, 1, 2],
    'shapes': [((4, 4), (4,)), ((5, 5), (5, 2))],
}))
@testing.fix_random()
@testing.with_requires('scipy')
class TestLUSolve(unittest.TestCase):

    @testing.for_dtypes('fdFD')
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def test_lu_solve(self, xp, scp, dtype):
        a_shape, b_shape = self.shapes
        A = testing.shaped_random(a_shape, xp, dtype=dtype)
        b = testing.shaped_random(b_shape, xp, dtype=dtype)
        lu = scp.linalg.lu_factor(A)
        return scp.linalg.lu_solve(lu, b, trans=self.trans)

    @requires_scipy_linalg_backend
    @testing.for_dtypes('fdFD')
    @testing.numpy_cupy_allclose(atol=1e-5)
    def test_lu_solve_backend(self, xp, dtype):
        a_shape, b_shape = self.shapes
        A = testing.shaped_random(a_shape, xp, dtype=dtype)
        b = testing.shaped_random(b_shape, xp, dtype=dtype)
        if xp is numpy:
            lu = scipy.linalg.lu_factor(A)
            backend = 'scipy'
        else:
            lu = cupyx.scipy.linalg.lu_factor(A)
            backend = cupyx.scipy.linalg
        with scipy.linalg.set_backend(backend):
            out = scipy.linalg.lu_solve(lu, b, trans=self.trans)
        return out



import unittest
import numpy

import pytest

import cupy
from cupyx import jit
from cupy import testing


class TestDeviceFunction(unittest.TestCase):

    def test_device_function(self):
        @jit.rawkernel()
        def f(x, y, z):
            tid = jit.threadIdx.x
            z[tid] = g(x[tid], y[tid]) + x[tid] + y[tid]

        @jit.rawkernel(device=True)
        def g(x, y):
            x += 1
            y += 1
            return x + y

        x = testing.shaped_random((30,), dtype=numpy.int32, seed=0)
        y = testing.shaped_random((30,), dtype=numpy.int32, seed=1)
        z = testing.shaped_random((30,), dtype=numpy.int32, seed=2)
        f((1,), (30,), (x, y, z))
        assert (z == (x + y + 1) * 2).all()

    def test_device_function_duplicated_names(self):
        @jit.rawkernel()
        def f(x, y, z):
            tid = jit.threadIdx.x
            z[tid] = g(10)(x[tid], y[tid])
            z[tid] += g(20)(x[tid], y[tid])
            z[tid] += g(30)(x[tid], y[tid])

        def g(const):
            @jit.rawkernel(device=True)
            def f(x, y):
                return x + y + const
            return f

        x = testing.shaped_random((30,), dtype=numpy.int32, seed=0)
        y = testing.shaped_random((30,), dtype=numpy.int32, seed=1)
        z = testing.shaped_random((30,), dtype=numpy.int32, seed=2)
        f((1,), (30,), (x, y, z))
        assert (z == (x + y) * 3 + 60).all()

    def test_device_function_recursive(self):
        @jit.rawkernel()
        def f(x, y, z):
            tid = jit.threadIdx.x
            z[tid] = g(x[tid], y[tid])

        @jit.rawkernel(device=True)
        def g(x, y):
            return x + y + g(x, y)

        x = testing.shaped_random((30,), dtype=numpy.int32, seed=0)
        y = testing.shaped_random((30,), dtype=numpy.int32, seed=1)
        z = testing.shaped_random((30,), dtype=numpy.int32, seed=2)
        with pytest.raises(ValueError):
            f((1,), (30,), (x, y, z))

    def test_device_function_template_recursion(self):
        @jit.rawkernel()
        def f(x, y):
            tid = jit.threadIdx.x
            y[tid] = x[tid]
            jit.syncthreads()
            g(1)(y)

        def g(step):
            @jit.rawkernel(device=True)
            def f(x):
                if step < 256:
                    tid = jit.threadIdx.x
                    if tid % (step * 2) == 0:
                        x[tid] += x[tid + step]
                        jit.syncthreads()
                        g(step * 2)(x)
            return f

        x = testing.shaped_random((256,), dtype=numpy.int32, seed=0)
        y = testing.shaped_random((256,), dtype=numpy.int32, seed=1)
        f((1,), (256,), (x, y))
        assert y[0] == x.sum()

    def test_device_function_called_once(self):
        @jit.rawkernel(device=True)
        def g(x):
            x[0] += 1
            return 1

        @jit.rawkernel()
        def f(x):
            x[g(x)] += 1

        x = cupy.array([0, 0])
        f((1,), (1,), (x,))
        testing.assert_array_equal(x, [1, 1])
